/*
Create plots of Digraphs in dot format.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <float.h>

#include "reporter.h"
#include "graph.h"
#include "privategraph.h"
#include "plotgraph.h"

#define TRUE 1
#define FALSE 0
#define MAX_LABEL_LENGTH 100

void plotDigraphHeader(Graph G) {
  char *author;			/* author's name */
  time_t t;			/* time of day */
  char *timestring;
  char graphLabel[MAX_LABEL_LENGTH+1];		/* graph label string */
  char *attributes = "size=\"7.5,10\";\ncenter=true;\n";
  int ifSort;			/* topological sort has been done */
  int ifPath;			/* unweighted shortest path has been done */

  author = getGraphAuthor();

  ifSort = (G->topNum != NULL);
  ifPath = (G->uPathTable != NULL) && (G->uPathStart != NULL);

  if (ifSort) {
    printf("/* Graph has a topological sort. */\n");
  } else {
    printf("/* Graph has no topological sort (or it has a cycle). */\n");
  }

  if (ifPath) {
    printf("/* Graph has shortest path table.  Start node is %s. */\n",G->uPathStart->name);
  } else {
    printf("/* Graph has no shortest path table. */\n");
  }

  t = time(NULL);
  timestring = ctime(&t);
  timestring[strlen(timestring)-1] = '\0';

  graphLabel[0] = '\0';
  strncat(graphLabel,G->title,MAX_LABEL_LENGTH-strlen(graphLabel));
  strncat(graphLabel,"\\n",MAX_LABEL_LENGTH-strlen(graphLabel));
  strncat(graphLabel,timestring,MAX_LABEL_LENGTH-strlen(graphLabel));
  strncat(graphLabel," - ",MAX_LABEL_LENGTH-strlen(graphLabel));
  strncat(graphLabel,author,MAX_LABEL_LENGTH-strlen(graphLabel));

  printf("digraph %s {\n","G");
  printf("label=\"%s\";\n",graphLabel);
  printf("%s",attributes);
}

/*

Print a line that defines a node in the digraph.  The node name is
the address of the object, the node label contains the actual name.

*/
void plotDigraphNodes(Graph G) {
  int i;
  struct Node *nodeP;		/* pointer to the current node */
  int ifSort;			/* topological sort has been done */
  int ifPath;			/* unweighted shortest path has been done */

  ifSort = (G->topNum != NULL);
  ifPath = (G->uPathTable != NULL) && (G->uPathStart != NULL);

  /* Print a line for each node. */

  for (i=0; i<G->nodeCount; i++) {
    nodeP = G->node[i];
    printf("n%p [label = \"%s\" ];",nodeP,nodeP->name);
    printf("\t/* ");
    printf("idx: %i, indegree: %i",nodeP->index,nodeP->indegree);
    if (ifSort) printf(", tsort: %i",G->topNum[i]);
    if (ifPath) {
      if (G->uPathTable[i].dist < DBL_MAX) {
	printf(", dist: %1.0f",G->uPathTable[i].dist);
	if (G->uPathTable[i].aP != NULL) {
	  printf(" via %s",G->uPathTable[i].aP->name);
	} else {
	  printf(" (start node)");
	}
      } else {
	printf(", dist: inf");
      }
    }
    printf(" */\n");
  }
}

/*

Print the lines that define edges in the digraph.  The node name is
the address of the object. 

*/
void plotDigraphEdges(Graph G) {
  int i;
  struct Node *nodeP;
  struct Edge *edgeP;
  Position pos;
  int ifPath;			/* unweighted shortest path has been done */
  int att;			/* at least one attribute field written */

  ifPath = (G->uPathTable != NULL) && (G->uPathStart != NULL);

  for (i=0; i<G->nodeCount; i++) {
    nodeP = G->node[i];
    pos = GetFirstListEntry(G->adj[i]);
    while (pos != NULL) {
      edgeP = (struct Edge *)GetListEntryElement(pos);
      printf("\"n%p\" -> \"n%p\" ",nodeP,edgeP->bP);
      att = FALSE;
      if (edgeP->weight > 0.0) {
	printf("%c",(att)?',':'[');
	att = TRUE;
	printf(" label = \"%1.0f\"",edgeP->weight);
      }
      if (ifPath) {
	printf("%c",(att)?',':'[');
	att = TRUE;
	if (G->uPathTable[edgeP->bP->index].aP == edgeP->aP) {
	  printf(" style = bold");
	} else {
	  printf(" style = dotted");
	}
      }
      if (att) printf(" ]");
      printf(";\n");
      pos = AdvanceListEntry(pos);
    }
  }
}

void plotDigraphFooter(Graph G) {
  printf("} /* %i nodes, %i edges */\n",G->nodeCount,G->edgeCount);
}
