#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "reporter.h"
#include "graph.h"
#include "privategraph.h"

#define MAX_LINE_LENGTH 400
#define MAX_NAME_LENGTH 100
#define MAX_NODE_COUNT 5000	/* nodes */
#define MAX_EDGE_COUNT 5000	/* edges */

/* configuration file section numbers */

#define sTITLE 0
#define sNODES 1
#define sEDGES 2
#define sEND 99

/*

Read in a graph definition file.  Return a pointer to a
Graph struct, which contains pointers to arrays of pointers
to Nodes and Edges, etc.

A graph definition file comprises three sections.

[TITLE]
A one line text title
[NODES]
<node name>
<node name>
[EDGES]
<edge name> <node name A> <node name B> <weight>
<edge name> <node name A> <node name B> <weight>
[END]

The <weight> entry is optional and defaults to zero.

If there is a memory allocation error, this procedure halts
immediately.

*/

static HashIndex hashNode(void *a,int size);
static int compareNode(void *a, void *b);
 
Graph readGraph(FILE *src) {

  char lineFmt[100];		/* format string for reading input line */
  char line[MAX_LINE_LENGTH+1]; /* an input line */
  char itemName[MAX_NAME_LENGTH+1];	/* an item name */
  char nodeA[MAX_NAME_LENGTH+1]; /* one end of an edge */
  char nodeB[MAX_NAME_LENGTH+1]; /* the other end of an edge */

  char nodeFmt[100];	/* formats for decoding each line */
  char edgeFmt[100];

  int status;			/* status return from fscanf */
  int itemStatus;		/* status return from sscanf */
  int section;			/* section that we are currently reading */

  Graph net;			/* pointer to GraphHeader object */

  struct Node *jnc;		/* pointer to a Node object */
  struct Node **jncP;		/* pointer to array of pointers */
  int nodeCount;			/* number of nodes read */

  struct Edge *lnk;		/* pointer to an Edge object */
  struct Edge **lnkP;		/* pointer to array of pointers */
  int edgeCount;			/* number of edges read */

  double weight;		/* edge attribute */
  int i;			/* loop control */
  int idx;			/* array subscript */

  net = malloc(sizeof(struct GraphHeader));
  FatalErrorMemory(net);

  net->title = NULL;
  net->node = NULL;
  net->edge = NULL;
  net->topNum = NULL;
  net->uPathTable = NULL;
  net->uPathStart = NULL;

  jncP = malloc(MAX_NODE_COUNT*sizeof(struct Node *));
  FatalErrorMemory(jncP);
  nodeCount = 0;

  lnkP = malloc(MAX_EDGE_COUNT*sizeof(struct Edge *));
  FatalErrorMemory(lnkP);
  edgeCount = 0;

  sprintf(lineFmt,"%%%i[^\n\r]%%*[\n\r]",MAX_LINE_LENGTH);
  sprintf(nodeFmt,"%%%is",MAX_NAME_LENGTH);
  sprintf(edgeFmt,"%%%is%%%is%%%is%%lf",
	  MAX_NAME_LENGTH,MAX_NAME_LENGTH,MAX_NAME_LENGTH);

  section = -1;

  /* read the configuration file, line by line */

  while ((status = fscanf(src,lineFmt,line)) == 1) {
    if (line[0] == '[') {	/* this line is start of a section */
      if (strcmp(line,"[TITLE]") == 0) 
	section = sTITLE;
      else if (strcmp(line,"[NODES]") == 0)
	section = sNODES;
      else if (strcmp(line,"[EDGES]") == 0)
	section = sEDGES;
      else if (strcmp(line,"[END]") == 0)
	section = sEND;
      else
	section = -1;
    } else if (section >= 0)	/* this line is within a section */

      switch (section) {
      case sTITLE:		/* keep only the first line of title */
	if (net->title == NULL) {
	  net->title = malloc((strlen(line)+1)*sizeof(char));
	  FatalErrorMemory(net->title);
	  strcpy(net->title,line);
	}
	break;

      case sNODES:		/* store info about each node */
	itemStatus = sscanf(line,nodeFmt,itemName);
	if (itemStatus >= 1 && itemName[0] != ';') {

	  nodeCount++;
	  FatalErrorBound(nodeCount,MAX_NODE_COUNT);

	  jnc = malloc(sizeof(struct Node));
	  FatalErrorMemory(jnc);

	  jnc->name = malloc((strlen(itemName)+1)*sizeof(char));
	  FatalErrorMemory(jnc->name);
	  strcpy(jnc->name,itemName);

	  jnc->indegree = 0;
	  jnc->index = nodeCount-1;
	  jncP[jnc->index] = jnc;
	}
	break;

      case sEDGES:		/* store info about each edge */
	itemStatus = sscanf(line,edgeFmt,itemName,nodeA,nodeB,&weight);
	if (itemStatus >= 3 && itemName[0] != ';') {
	  lnk = malloc(sizeof(struct Edge));
	  FatalErrorMemory(lnk);

	  lnk->name = malloc((strlen(itemName)+1)*sizeof(char));
	  FatalErrorMemory(lnk->name);
	  strcpy(lnk->name,itemName);

	  lnk->nodeA = malloc((strlen(nodeA)+1)*sizeof(char));
	  FatalErrorMemory(lnk->nodeA);
	  strcpy(lnk->nodeA,nodeA);

	  lnk->nodeB = malloc((strlen(nodeB)+1)*sizeof(char));
	  FatalErrorMemory(lnk->nodeB);
	  strcpy(lnk->nodeB,nodeB);

	  lnk->weight = (itemStatus > 3) ? weight : 0.0;

	  edgeCount++;
	  FatalErrorBound(edgeCount,MAX_EDGE_COUNT);
	  lnkP[edgeCount-1] = lnk;
	}
	break;

      case sEND:
	break;
      }
  }

  net->node = realloc(jncP,nodeCount*sizeof(struct Node *));
  FatalErrorMemory(net->node);
  net->nodeCount = nodeCount;

  net->edge = realloc(lnkP,edgeCount*sizeof(struct Edge *));
  FatalErrorMemory(net->edge);
  net->edgeCount = edgeCount;

  /* build a hash table to go from Node name to Node pointer */

  net->hash = CreateHashTable(nodeCount,hashNode,compareNode);
  for (i=0; i<nodeCount; i++) {
    InsertHashEntry(net->hash,jncP[i]);
  }

  /* use the hash table to store Node pointers in the Edges */

  for (i=0; i<edgeCount; i++) {

    lnkP[i]->aP = FindNodeByName(net,lnkP[i]->nodeA);
    FatalErrorObjectNotFound(lnkP[i]->aP,lnkP[i]->nodeA);

    lnkP[i]->bP = FindNodeByName(net,lnkP[i]->nodeB);
    FatalErrorObjectNotFound(lnkP[i]->bP,lnkP[i]->nodeB);
  }

  /*
    Create the adjacency lists.  
  */

  net->adj = CreateListArray(net->nodeCount);
  FatalErrorMemory(net->adj);

  for (i=0; i<net->edgeCount; i++) {
    (lnkP[i]->bP)->indegree++;
    idx = (lnkP[i]->aP)->index;
    InsertListEntry((net->adj)[idx],(net->adj)[idx],lnkP[i]);
  }

  return net;
}

/*-------------------------------------------------------------------*/

struct Node *FindNodeByName(Graph G,char *name) {
  Position p;
  struct Node j;
  j.name = name;
  p = FindHashEntry(G->hash,&j);
  return GetHashEntryElement(p);
}

/*-------------------------------------------------------------------*/

/*
  hashNode considers a character string to be a very large number
  expressed as a series of base 256 digits.  It calculates the remainder
  when that number is divided by the table size.
*/
static HashIndex hashNode(void *e,int size) {
  int dividend;
  int remainder;
  int radix;
  char *name;
  name = ((struct Node *)e)->name;
  radix = 256;
  remainder = 0;
  while (*name != '\0') {
    dividend = (radix * remainder) + *name;
    remainder = dividend % size;
    name++;
  }
  return remainder;
}

/*-------------------------------------------------------------------*/

static int compareNode(void *a, void *b) {
  char *nameA;
  char *nameB;
  nameA = ((struct Node *)a)->name;
  nameB = ((struct Node *)b)->name;
  return strcmp(nameA,nameB);
}
